//**************************************************************************//
//**                                                                      **//
//**                           NesForm                                    **//
//**                                                                      **//
//**   Copyright (C) Dott. Cosimo De Michele 1998                         **//
//**         You are not allowed to distribuite this software             **//
//**         commercially. Please, notify me, if you make any             **//
//**         changes to this file.                                        **//
//**                                                                      **//
//**************************************************************************//

unit NesForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ComCtrls, StdCtrls, ExtCtrls, DXClass, DXDraws, DIB, DXSprite,
  DXInput, VersionInfo, MF6502, MidiOut;

type
  TMemory = array[0..maxint-1] of BYTE;
  PMemory = ^TMemory;

  TPalType = array[0..63, 1..3] of byte;

  TSPR = record
           y, no, atr, x : byte;
         end;

const
  NES_SCREEN_WIDTH  = 256;
  NES_SCREEN_HEIGHT = 240;
  MINLINE           = 8;
  MAXLINE           = 231; //238;
  SCANLINEPERFRAME  = 244;
  SNDPERIOD         = 244;
  NO_DATA           = $FF;

  ROM_MIRROR  = 1;
  ROM_WRAM    = 2;
  ROM_TRAINER = 4;
  ROM_FOUR    = 8;

  PPU_InVBlank    = $80;
  PPU_SpHit       = $40;
  PPU_SpMax       = $20;
  PPU_WriteIgnore = $10;

  JOY_A      = $01;
  JOY_B      = $02;
  JOY_SELECT = $04;
  JOY_START  = $08;
  JOY_UP     = $10;
  JOY_DOWN   = $20;
  JOY_LEFT   = $40;
  JOY_RIGHT  = $80;

  V_FLIP     = $80;
  H_FLIP     = $40;
  SP_BACK    = $20;
  PSG_BASE   = 111860;

  MRUCount   = 5;
  RegKey     = 'Software\Karakiri\WinNES';
  midiNotePlay = $90; { 144 }
  midiNoteSilence = $80;
  midiChangeInstrument = $C0; { 192 }

  Palette : TPalType =
    {Alex Krasivsky's Palette}
    (
    ($68,$68,$68),($00,$40,$80),($00,$00,$80),($40,$00,$80),
    ($80,$00,$80),($80,$00,$40),($80,$00,$00),($55,$00,$00),
    ($40,$40,$00),($33,$50,$00),($00,$50,$00),($00,$50,$00),
    ($00,$40,$40),($00,$00,$00),($00,$00,$00),($00,$00,$00),
    ($98,$98,$98),($00,$80,$C0),($40,$40,$C0),($80,$00,$C0),
    ($C0,$00,$C0),($C0,$00,$80),($C0,$20,$20),($C0,$40,$00),
    ($80,$80,$00),($55,$80,$00),($00,$80,$00),($33,$80,$33),
    ($00,$80,$80),($00,$00,$00),($00,$00,$00),($00,$00,$00),
    ($D0,$D0,$D0),($40,$C0,$FF),($80,$80,$FF),($C0,$40,$FF),
    ($FF,$00,$FF),($FF,$40,$C0),($FF,$50,$50),($FF,$80,$40),
    ($C0,$C0,$00),($80,$C0,$00),($00,$C0,$00),($55,$C0,$55),
    ($00,$C0,$C0),($33,$33,$33),($00,$00,$00),($00,$00,$00),
    ($FF,$FF,$FF),($80,$FF,$FF),($C0,$C0,$FF),($FF,$80,$FF),
    ($FF,$40,$FF),($FF,$80,$FF),($FF,$80,$80),($FF,$C0,$80),
    ($FF,$FF,$40),($C0,$FF,$40),($40,$FF,$40),($AA,$FF,$AA),
    ($40,$FF,$FF),($99,$99,$99),($00,$00,$00),($00,$00,$00)
    );

type
  TRomHeader = record
    RomType : array[1..3] of char;
    Ctrl,
    PRGROM,
    CHRRAM,
    ROMctrl1,
    ROMctrl2 : char;
    Expansion : array[0..7] of char;
  end;

  TfrmNES = class(TDXForm)
    MainMenu: TMainMenu;
    mnuFile: TMenuItem;
    mnuLoadROM: TMenuItem;
    mnuFreeROM: TMenuItem;
    mnuRom: TMenuItem;
    N2: TMenuItem;
    mnuExit: TMenuItem;
    Help1: TMenuItem;
    mnuAbout: TMenuItem;
    StatusBar: TStatusBar;
    OpenDialog: TOpenDialog;
    N3: TMenuItem;
    Terminal: TMemo;
    mnuCPU: TMenuItem;
    mnuReset: TMenuItem;
    mnuPause: TMenuItem;
    mnuSavesnapshot: TMenuItem;
    MRUSeparator: TMenuItem;
    MRU1: TMenuItem;
    MRU2: TMenuItem;
    MRU3: TMenuItem;
    MRU4: TMenuItem;
    MRU5: TMenuItem;
    pmuLoadRom: TMenuItem;
    pmuFreeROM: TMenuItem;
    N1: TMenuItem;
    pmuReset: TMenuItem;
    pmuPause: TMenuItem;
    N5: TMenuItem;
    mnuSettings: TMenuItem;
    N4: TMenuItem;
    mnuDebug: TMenuItem;
    DXDIB: TDXDIB;
    DXInput1: TDXInput;
    DXInput2: TDXInput;
    VersionResource: TVersionResource;
    SaveDialog: TSaveDialog;
    DXDraw: TDXDraw;
    DXImageList: TDXImageList;
    PopupMenu: TPopupMenu;
    MidiOutput: TMidiOutput;
    DXImageList1: TDXImageList;
    procedure mnuLoadROMClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mnuExitClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure mnuPauseClick(Sender: TObject);
    procedure mnuResetClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure mnuRomClick(Sender: TObject);
    procedure mnuAboutClick(Sender: TObject);
    procedure mnuFreeROMClick(Sender: TObject);
    procedure mnuSettingClick(Sender: TObject);
    procedure mnuSavesnapshotClick(Sender: TObject);
    procedure MRUClick(Sender: TObject);
    procedure mnuFileClick(Sender: TObject);
    procedure mnuCPUClick(Sender: TObject);
    procedure StatusBarDrawPanel(StatusBar: TStatusBar;
      Panel: TStatusPanel; const Rect: TRect);
    procedure mnuDebugClick(Sender: TObject);
  private
    CPUPeriod,
    UPeriod,
    SndCount,
    HitY : integer;
    ScreenWidth,
    ScreenHeight,
    CPUinstr,
    ScanLine,
    LastTimerTick,
    CurrentFrame : integer;
    UseDirectX: boolean;
    { Sound }
    SoundOn : boolean;
    { Cartridge setting }
    RomFilename,
    RomPath     : string;
    RomHeader   : TRomHeader;
    { Joypad 1 & 2 }
    Joypad1,
    Joypad2,
    Joypad1Read,
    Joypad2Read  : byte;
    { PPU register }
    PPU_PalChange,
    PPUADDR_FirstRead,
    PPUSCROLL_Toggle : boolean;
    PPUSCROLL_X1,
    PPUSCROLL_Y1,
    PPUSCROLL_X2,
    PPUSCROLL_Y2 : byte;
    MRUList : TStringList;
    { PSG register }
    SndFreq : array[0..3] of boolean;
    SndFreqVal : array[0..3] of word;
    SndLen : array[0..3] of integer;
    KeyOn : array[0..3] of integer;
    { }
    procedure RunMachine(Cartridge : string);
    function  LoadNES(filename : string): boolean;
    function  InitNES: boolean;
    procedure TrashMachine;
    procedure IOWrite(A : word; V : byte);
    function  IORead(A : word): byte;
    procedure CheckKb;
    procedure RefreshLine (Y1 : integer);
    procedure CheckSpriteHit;
    procedure RefreshSprite(Y1 : integer);
    function  InitSound: boolean;
    procedure InitScreen;
    procedure DrawScreen;
    procedure TrashSound;
    procedure TrashScreen;
    procedure GoFullScreen;
    procedure ReadConfig;
    procedure WriteConfig;
    procedure SetFreq(ch, freq : WORD);
    procedure SetVolume(ch, vol : integer);
    procedure PlayLogo;
  public
    PPUSCROLL_Change: boolean;
    UCount: integer;
    {}
    cpu : M6502;
    NESisRunning : boolean;
    Page,
    VPage : array [0..8] of pMemory;
    { Cartridge ROM }
    ROM,
    VROM  : pMemory;
    VAddr : word;
    { PPU }
    PPU    : array[0..3] of byte;
    VRAM   : array[0..$4000] of byte;
    SRAM   : array[0..$0100] of byte;
    WRAM   : array[0..$1FFF] of byte;
    { PSG }
    PSG    : array [0..$17] of byte;
    SndON  : array[0..3] of boolean;
    SndVolume : integer;

    ROMSize,
    VROMSize,
    ROMType,
    MMCType,
    Mirroring,
    MirrorXor  : integer;
    FrameRate  : integer;
    FourScreen : byte;
    ViewFreq : boolean;
    { NES Keyboard }
    BUTTON1_UP,
    BUTTON1_DOWN,
    BUTTON1_LEFT,
    BUTTON1_RIGHT,
    BUTTON1_A,
    BUTTON1_B,
    BUTTON1_START,
    BUTTON1_SELECT,
    BUTTON2_UP,
    BUTTON2_DOWN,
    BUTTON2_LEFT,
    BUTTON2_RIGHT,
    BUTTON2_A,
    BUTTON2_B,
    BUTTON2_START,
    BUTTON2_SELECT,
    WIN_MODE : integer;
    BUTTON1_JOY_ENABLED,
    BUTTON2_JOY_ENABLED : boolean;
    SCREEN_MODE : string;
    { Central Memory }
    RAM : array [0..$800] of byte;
    function MemoryReadOpcode(Addr : word) : byte;
    function MemoryReadByte(Addr : word) : byte;
    procedure MemoryWrite (Addr : word; value : byte);
    function cpuLoop : byte;
    procedure ResetNES;
  end;

var
  frmNES: TfrmNES;
  Version : string;

implementation

uses
  Registry,
  RomPropertyForm,
  AboutForm,
  SettingForm,
  Mapper,
  Math, DebugForm;

{$R *.DFM}

procedure AssignKey(var KeyAssignList: TKeyAssignList; State: TDXInputState;
  const Keys: array of Integer);
var
  i, i2: Integer;
  KeyAssign: PKeyAssign;
begin
  KeyAssign := @KeyAssignList[State];
  FillChar(KeyAssign^, SizeOf(TKeyAssign), 0);
  i2 := 0;
  for i:=LOW(Keys) to HIGH(Keys) do
  begin
    if i2<3 then
      KeyAssign^[i2] := Keys[i]
    else
      Exit;
    Inc(i2);
  end;
end; { AssignKey }

procedure TfrmNES.SetFreq(ch, freq : WORD);
var
  nota : byte;
begin
  if not(SoundOn) or (freq<10) or (freq>14000) or not(sndON[ch]) then
    Exit;
  nota := Trunc(12 * (Log2(freq*32/440))) + 9;
  { setto la nota }
  MidiOutput.PutShort(midiNotePlay + ch, nota, 64);
  MidiOutput.PutShort(midiNoteSilence + ch, KeyOn[ch], 0);
  KeyOn[ch] := nota;
end; { SetFreq }

procedure TfrmNES.SetVolume(ch, vol : integer);
begin
  if not(SoundOn) or not(sndON[ch]) then
    Exit;
  if vol=0 then
    MidiOutput.PutShort(midiNoteSilence + ch, KeyOn[ch], 0)
  else
    MidiOutput.SetVolume(vol*6553, vol*6553);
end; { SetVolume }

function  TfrmNES.InitSound: boolean;
var
  i: integer;
begin
  for i:=0 to 3 do
    KeyOn[i] := 0;
  MidiOutput.DeviceID := -1;
  Result := true;
  try
    MidiOutput.Open;
    MidiOutput.PutShort(midiChangeInstrument + 0, 81, 0);
    MidiOutput.PutShort(midiChangeInstrument + 1, 81, 0);
    MidiOutput.PutShort(midiChangeInstrument + 2, 81, 0);
    MidiOutput.PutShort(midiChangeInstrument + 3, 119, 0);
  except
    Result := false;
  end;
end; { InitSound }

procedure TfrmNES.TrashSound;
begin
  { sound }
  MidiOutput.Close;
  SoundOn := false;
end; { TrashSound }

procedure TfrmNES.RunMachine(Cartridge : string);
begin
  mnuFreeROM.Click;
  if LoadNES(Cartridge) then
  begin
    if InitNES then
    begin
      mnuRom.enabled := true;
      mnuFreeROM.enabled := true;
      mnuDebug.enabled := true;
      mnuSaveSnapshot.enabled := true;
      { reset the NES machine }
      ResetNes;
      Terminal.Lines.Add('6502 engine started.');
      { reset the sound }
      SoundOn := InitSound;
      { run the NES machine }
      InitScreen;
      NESisRunning := true;
      StatusBar.Panels[0].Text := UpperCase(RomFilename);
      StatusBar.Refresh;
      Run6502(cpu);
    end;
  end;
end; { RunMachine }

procedure TfrmNES.mnuLoadROMClick(Sender: TObject);
begin
  if doFullScreen in DXDraw.Options then
    mnuFreeROM.Click;
  if OpenDialog.Execute then
  begin
    { update MRU list }
    if MRUList.indexOf(OpenDialog.Filename)=-1 then
    begin
      MRUList.Insert(0, OpenDialog.Filename);
      if MRUList.Count>MRUCount then
        MRUList.Delete(MRUCount);
    end;
    { run the NES }
    RunMachine(OpenDialog.Filename);
  end;
end; { mnuLoadROMClick }

procedure TfrmNES.ResetNES;
var
  i : integer;
begin
  UPeriod := 0;
  UCount  := 0;
  { PSG register }
  for i := 0 to 14 do
    PSG[i] := 0;
  { Joypads }
  Joypad1 := 0;
  Joypad2 := 0;
  Joypad1Read := 0;
  Joypad2Read := 0;
  { PPU register }
  PPU[0] := 0;
  PPU[1] := 0;
  PPU[2] := 0;
  PPU[3] := 0;
  VAddr  := 0;
  PPUSCROLL_X1 := 0;
  PPUSCROLL_Y1 := 0;
  PPUSCROLL_X2 := 0;
  PPUSCROLL_Y2 := 0;
  { Screen }
  PPUADDR_FirstRead  := true;
  PPUSCROLL_Toggle := false;
  scanline := 0;
  Terminal.Lines.Add('Resetting 6502 engine ...');
  { Reset cpu }
  CPUinstr := 0;
  Reset6502(cpu);
end; { ResetNES }

procedure TfrmNES.TrashMachine;
var
  i : integer;
  BatteryName : string;
  BatteryFile : file;
begin
  { fermo la cpu }
  NESisRunning := false;
  Terminal.Lines.Add('6502 engine stopped.');
  if (ROMType and ROM_WRAM)<>0 then
  begin
    { Save backed memory }
    BatteryName := RomPath + Copy(RomFilename, 1, Pos('.', RomFilename)) + 'sav';
    Terminal.Lines.Add('Saving backed memory file "' + ExtractFilename(BatteryName) + '".');
    AssignFile(BatteryFile, BatteryName);
    ReWrite(BatteryFile, 1);
    BlockWrite(BatteryFile, WRAM, SizeOf(WRAM), i);
    CloseFile(BatteryFile);
  end;
  { azzero memoria }
  FillChar(VRAM, SizeOf(VRAM), $00);
  FillChar(SRAM, SizeOf(SRAM), $00);
  FillChar(WRAM, SizeOf(WRAM), $00);
  FillChar(RAM, SizeOf(RAM), $00);
  for i:=Low(Page) to High(Page) do
    Page[i] := nil;
  for i:=Low(VPage) to High(VPage) do
    VPage[i] := nil;
  if Assigned(ROM) then
  begin
    FreeMem(ROM);
    ROM := nil;
  end;
  if Assigned(VROM) then
  begin
    FreeMem(VROM);
    VROM := nil;
  end;
  RomFilename := '';
  RomPath     := '';
  { sound }
  if SoundOn then
    TrashSound;
  { ripristino lo schermo }
  TrashScreen;
  Terminal.Lines.Add('Cartridge discharged.');
  StatusBar.Panels[0].Text := '';
end; { TrashMachine }

function TfrmNES.InitNES: boolean;
var
  j : integer;
  BatteryName : string;
  BatteryFile : file;
begin
  Page[3] := @WRAM;
  if (ROMType and ROM_WRAM)<>0 then
  begin
    { Loading of backed memory }
    Terminal.Lines.Add('A backed memory has been found.');
    BatteryName := RomPath + Copy(RomFilename, 1, Pos('.', RomFilename)) + 'sav';
    if FileExists(BatteryName) then
    begin
      Terminal.Lines.Add('Loading backed memory file "' + ExtractFilename(BatteryName) + '".');
      AssignFile(BatteryFile, BatteryName);
      Reset(BatteryFile, 1);
      BlockRead(BatteryFile, WRAM, SizeOf(WRAM), j);
      CloseFile(BatteryFile);
    end;
  end;
  Page[0] := @RAM;
  Result := MMCInit(MMCType);
end; { InitNES }

function TfrmNES.LoadNES(filename : string) : boolean;
var
  F : file;
  i, j,
  PRGBytes,
  CHRBytes,
  CartSize : integer;
begin
  frmNES.Terminal.Lines.Add('');
  RomFilename := ExtractFilename(filename);
  RomPath     := ExtractFilePath(filename);
  Terminal.Lines.Add('Loading Cartridge : "' + RomFileName +'".');
  AssignFile(F, filename);
  Reset(F, 1);
  { Header }
  BlockRead(F, ROMHeader, SizeOf(ROMHeader), i);
  if (ROMHeader.RomType <> 'NES') then
  begin
   Terminal.Lines.Add('Unsupported file format.');
   Result := false;
   Exit;
  end;
  ROMSize  := Ord(ROMHeader.PRGROM);
  VROMSize := Ord(ROMHeader.CHRRAM);
  ROMType  := Ord(ROMHeader.ROMctrl1);
  MMCType  := (Ord(ROMHeader.ROMctrl1) shr 4);
  CartSize := ROMSize shl 4 + VROMSize * 8;
  if ROMHeader.Expansion[0]=#0 then
    MMCType := MMCType or (Ord(ROMHeader.ROMctrl2) and $F0);

  Terminal.Lines.Add(format('iNES #%d file format found.', [MMCType]));
  Terminal.Lines.Add(format('[%d] of PRG Rom', [ROMSize]));
  Terminal.Lines.Add(format('[%d] of CHR Ram', [VROMSize]));
  Terminal.Lines.Add(format('Size of the cartridge : [%d Kb]', [CartSize]));

  Mirroring := (Ord(ROMHeader.ROMctrl1) and ROM_MIRROR);
  MirrorXor := $400 shl Mirroring; { Horizontal : $400; Vertical : $800) }
  FourScreen := (Ord(ROMHeader.ROMctrl1) and ROM_FOUR);
  { dimensionamento aree di memoria }
  PRGBytes := $4000 * ROMSize;
  CHRBytes := $2000 * VROMSize;
  GetMem(ROM, PRGBytes + 1);
  GetMem(VROM, CHRBytes + 1);
  { Caricamento della ROM }
  BlockRead(F, ROM^, PRGBytes, j);
  { Caricamento della VROM }
  BlockRead(F, VROM^, CHRbytes, j);
  CloseFile(F);
  Result := true;
end; { LoadNES }

procedure TfrmNES.CheckKb;
begin
  Joypad1 := 0;
  { Joypad1 }
  DXInput1.Update;
  { KEY A }
  if (isButton1 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_A;
  { KEY B }
  if (isButton2 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_B;
  { KEY START }
  if (isButton3 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_START;
  { KEY SELECT }
  if (isButton4 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_SELECT;
  { KEY UP }
  if (isUp in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_UP;
  { KEY DOWN }
  if (isDown in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_DOWN;
  { KEY LEFT }
  if (isLeft in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_LEFT;
  { KEY RIGHT }
  if (isRight in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_RIGHT;

  Joypad2 := 0;
  { Joypad2 }
  DXInput2.Update;
  { KEY A }
  if (isButton1 in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_A;
  { KEY B }
  if (isButton2 in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_B;
  { KEY START }
  if (isButton3 in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_START;
  { KEY SELECT }
  if (isButton4 in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_SELECT;
  { KEY UP }
  if (isUp in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_UP;
  { KEY DOWN }
  if (isDown in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_DOWN;
  { KEY LEFT }
  if (isLeft in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_LEFT;
  { KEY RIGHT }
  if (isRight in DXInput2.States) then
    Joypad2 := Joypad2 or JOY_RIGHT;
end; { CheckKb }

procedure TfrmNES.PlayLogo;
var
  Line,
  FAngle : integer;
begin
  DXImageList.Items.MakeColorTable;
  DXDraw.DefColorTable := DXImageList.Items.ColorTable;
  DXDraw.ColorTable := DXImageList.Items.ColorTable;
  DXDraw.UpdatePalette;
  DXImageList.Items[1].Restore;
  FAngle := 0;
  repeat
    if DXDraw.CanDraw then
    begin
      DXDraw.Surface.Fill(0);
      DXDraw.Surface.LoadFromGraphic(DXImageList.Items[0].Picture.Graphic);
      DXImageList.Items[1].DrawRotateAlpha(DXDraw.Surface, DXDraw.Surface.Width div 2,
        DXDraw.Surface.Height div 2, DXImageList.Items[1].Width, DXImageList.Items[1].Height,
        0, 0.5, 0.5, FAngle, Trunc(Cos256(FAngle)*126+127));
      DXDraw.Flip;
    end;
    inc(FAngle, 2);
  until FAngle>130;
  { Cancello la memoria video }
  DXDraw.Surface.Fill(0);
  for Line := MINLINE to MAXLINE do
    FillChar(DXDIB.DIB.ScanLine[Line]^, DXDIB.DIB.WidthBytes, 0);
end; { PlayLogo }

procedure TfrmNES.DrawScreen;
var
  r : double;
  x, y, i, h, t : longint;
  NESRGBQuads : TRGBQuads;
  pal: PLogPalette;
  hpal: HPALETTE;
begin
  if UseDirectX then
  begin
    if not DXDraw.CanDraw then exit;
    if (PPU_PalChange) then
    begin
      for i:=0 to 63 do
        NESRGBQuads[i] := RGBQuad(Palette[VRAM[$3F00 + i], 1], Palette[VRAM[$3F00 + i], 2], Palette[VRAM[$3F00 + i], 3]);
      { inizializzo buffer video }
      DXDIB.DIB.ColorTable := NESRGBQuads;
      DXDIB.DIB.UpdatePalette;
      DXDraw.ColorTable := DXDIB.DIB.ColorTable;
      DXDraw.UpdatePalette;
      PPU_PalChange := false;
    end;
    //DXDraw.Surface.Fill(0);
    //DXDraw.Surface.Assign(DXDIB.DIB);
    DXDraw.Surface.LoadFromGraphic(DXDIB.DIB);
    with DXDraw.Surface.Canvas do
    begin
      if ViewFreq then
      begin
        t := GetTickCount;
        r := 0;
        if (LastTimerTick<>0) and (t<>LastTimerTick) then
          r := (1000) / (t - LastTimerTick);
        LastTimerTick := t;
        Brush.Style := bsClear;
        Font.Color  := clWhite;
        Font.Size   := 6;
        h := NES_SCREEN_HEIGHT - TextHeight('F') + 1;
        Textout(0, h, format('FPS: %3.2f', [r]));
      end;
      Release;
    end;
    DXDraw.Flip;
  end
  else
  begin
    if (PPU_PalChange) then
    begin
      pal := nil;
      try
        GetMem(pal, sizeof(TLogPalette) + sizeof(TPaletteEntry) * 64);
        pal.palVersion := $300;
        pal.palNumEntries := 64;
        for i := 0 to 63 do
        begin
          pal.palPalEntry[i].peRed := Palette[VRAM[$3F00 + i], 1];
          pal.palPalEntry[i].peGreen := Palette[VRAM[$3F00 + i], 2];
          pal.palPalEntry[i].peBlue := Palette[VRAM[$3F00 + i], 3];
        end;
        hpal := CreatePalette(pal^);
        if hpal <> 0 then
          DXDIB.DIB.Palette := hpal;
      finally
        FreeMem(pal);
      end;
      PPU_PalChange := false;
    end;
    with Canvas do
    begin
      x := (ClientWidth - ScreenWidth) div 2;
      y := (ClientHeight - StatusBar.Height - ScreenHeight) div 2;
      Draw(x, y, DXDIB.DIB);
      if ViewFreq then
      begin
        t := GetTickCount;
        r := 0;
        if (LastTimerTick<>0) and (t<>LastTimerTick) then
          r := (1000) / (t - LastTimerTick);
        LastTimerTick := t;
        Brush.Style := bsClear;
        Font.Color  := clWhite;
        Font.Size   := 6;
        h := y + ScreenHeight - TextHeight('F') + 1;
        Textout(x, h, format('%3.2f FPS', [r]));
      end;
    end;
  end;
end; { DrawScreen }

procedure TfrmNES.RefreshLine(Y1 : integer);
var
  iPP : integer;
  C, nametbl, PP: pMemory;
  vadr,
  ntbl : word;
  x, y, X1, vofs : integer;
  no, atr, cc, c1, c2 : byte;
begin
  PP := DXDIB.DIB.ScanLine[Y1];
  if (PPU[1] and $8)=0 then { ScreenON }
    { Clear Line }
    FillChar(PP^, DXDIB.DIB.WidthBytes, 0)
  else
  begin { Draw Line }
    vofs := 0;
    if (PPU[0] and $10)<>0 then { BGAdrHI }
      vofs := $1000;
    y := Y1 + PPUSCROLL_Y1;
    inc(vofs, (y and 7));
    y := y shr 3;
    ntbl := $2000 + (PPU[0] and $3) * $400;
    iPP := -(PPUSCROLL_X1 and  7);
    x := PPUSCROLL_X1 shr 3;
    if (y>=30) then
    begin
      dec(y, 30);
      ntbl := ntbl xor $800;
    end;
    if x=32 then
      x := 0;
    nametbl := pMemory(pChar(@VRAM) + ntbl);

    for X1 := 0 to 32 do
    begin
      atr := nametbl[$03C0 + ( x shr 2 ) + ( y shr 2 ) * 8];
      cc := ((atr shr ((x and 2) + (y and 2) * 2)) and 3) * 4;
      no := nametbl[x + y * 32];
      vadr := no * 16 + vofs;
      C  := @VPage[vadr shr 10][vadr];
      c1 := ((C[0] shr 1) and $55) or (C[8] and $AA);
      c2 := (C[0] and $55) or ((C[8] shl 1) and $AA);
      if ((0 + ipp)>=0) and ((0 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[0 + iPP] := ((c1 shr 6) and 3) + cc;
      if ((1 + ipp)>=0) and ((1 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[1 + iPP] := ((c2 shr 6) and 3) + cc;
      if ((2 + ipp)>=0) and ((2 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[2 + iPP] := ((c1 shr 4) and 3) + cc;
      if ((3 + ipp)>=0) and ((3 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[3 + iPP] := ((c2 shr 4) and 3) + cc;
      if ((4 + ipp)>=0) and ((4 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[4 + iPP] := ((c1 shr 2) and 3) + cc;
      if ((5 + ipp)>=0) and ((5 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[5 + iPP] := ((c2 shr 2) and 3) + cc;
      if ((6 + ipp)>=0) and ((6 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[6 + iPP] := (c1 and 3) + cc;
      if ((7 + ipp)>=0) and ((7 + ipp)<DXDIB.DIB.WidthBytes) then
        PP[7 + iPP] := (c2 and 3) + cc;
      inc(x);
      if x=32 then
      begin
        x := 0;
        nametbl := pMemory(pChar(@VRAM) + (ntbl xor $400));
      end;
      inc(iPP, 8);
    end;
  end;
end; { RefreshLine }

procedure TfrmNES.RefreshSprite(Y1 : integer);
var
  n, vofs : integer;
  HH : byte;
  sprite : ^TSPR;
  PP: ^TByteArray;
  iPP, C : integer;
  x, y, no, atr, CC, J, c1, c2 : byte;
  h, t, incr, vadr : integer;
begin
  vofs := 0;
  if (PPU[0] and $20)<>0 then { Sprite 16 }
    HH := 16
  else
  begin
    HH := 8;
    if (PPU[0] and $08)<>0 then { Sprite pattern address }
      vofs := $1000;
  end;

  n := $00FC;
  repeat
    sprite := @SRAM[n];
    y := sprite.y;
    if not((y>=MAXLINE) or (y<(Y1-HH))) then
    begin
      x   := sprite.x;
      no  := sprite.no;
      atr := sprite.atr;
      cc  := (atr and $3) * 4 + 16;
      if (PPU[0] and $20)<>0 then { Sprite 16 }
        vadr := ((no and $1) shl 12) + (no shr 1)* 32
      else
        vadr := no * 16 + vofs;
      C := vadr;
      iPP  := x;
      incr := 1;
      if (atr and V_FLIP)<>0 then
      begin
        incr := -1;
        inc(C, 7);
        if (PPU[0] and $20)<>0 then { Sprite 16 }
          inc(C, 16);
      end;
      t := Y1 - y;
      if t>0 then
      begin
        inc(C, t*incr);
        iPP := x;
        if (t>8) then
          inc(C, incr*8);
      end
      else
        t := 0;
      h := MAXLINE - y;// + 1;
      if (h>HH) then
        h := HH;
      repeat
        PP  := DXDIB.DIB.ScanLine[y + t];
        if (t=8) then
          inc(C, incr*8);
        c1 := ((Ord(VPage[C shr 10][C]) shr 1) and $55) or (Ord(VPage[(C + 8) shr 10][C + 8]) and $AA);
        c2 := (Ord(VPage[C shr 10][C]) and $55) or ((Ord(VPage[(C + 8) shr 10][C + 8]) shl 1) and $AA);
        J := Ord(VPage[C shr 10][C]) or Ord(VPage[(C + 8) shr 10][C + 8]);
        if (J<>0) and ((atr and SP_BACK)<>0) then
        begin
          if (atr and H_FLIP)<>0 then
          begin
            if ((J and $80)<>0) and ((PP[iPP + 7] and 3)=0) and ((iPP + 7)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 7] := ((c1 shr 6) and 3) + cc;
            if ((J and $40)<>0) and ((PP[iPP + 6] and 3)=0) and ((iPP + 6)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 6] := ((c2 shr 6) and 3) + cc;
            if ((J and $20)<>0) and ((PP[iPP + 5] and 3)=0) and ((iPP + 5)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 5] := ((c1 shr 4) and 3) + cc;
            if ((J and $10)<>0) and ((PP[iPP + 4] and 3)=0) and ((iPP + 4)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 4] := ((c2 shr 4) and 3) + cc;
            if ((J and $8)<>0) and ((PP[iPP + 3] and 3)=0) and ((iPP + 3)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 3] := ((c1 shr 2) and 3) + cc;
            if ((J and $4)<>0) and ((PP[iPP + 2] and 3)=0) and ((iPP + 2)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 2] := ((c2 shr 2) and 3) + cc;
            if ((J and $2)<>0) and ((PP[iPP + 1] and 3)=0) and ((iPP + 1)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 1] := (c1 and 3) + cc;
            if ((J and $1)<>0) and ((PP[iPP + 0] and 3)=0) and ((iPP + 0)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 0] := (c2 and 3) + cc;
          end
          else
          begin
            if ((J and $80)<>0) and ((PP[iPP + 0] and 3)=0) and ((iPP + 0)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 0] := ((c1 shr 6) and 3) + cc;
            if ((J and $40)<>0) and ((PP[iPP + 1] and 3)=0) and ((iPP + 1)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 1] := ((c2 shr 6) and 3) + cc;
            if ((J and $20)<>0) and ((PP[iPP + 2] and 3)=0) and ((iPP + 2)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 2] := ((c1 shr 4) and 3) + cc;
            if ((J and $10)<>0) and ((PP[iPP + 3] and 3)=0) and ((iPP + 3)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 3] := ((c2 shr 4) and 3) + cc;
            if ((J and $8)<>0) and ((PP[iPP + 4] and 3)=0) and ((iPP + 4)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 4] := ((c1 shr 2) and 3) + cc;
            if ((J and $4)<>0) and ((PP[iPP + 5] and 3)=0) and ((iPP + 5)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 5] := ((c2 shr 2) and 3) + cc;
            if ((J and $2)<>0) and ((PP[iPP + 6] and 3)=0) and ((iPP + 6)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 6] := (c1 and 3) + cc;
            if ((J and $1)<>0) and ((PP[iPP + 7] and 3)=0) and ((iPP + 7)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 7] := (c2 and 3) + cc;
          end;
        end
        else if J<>0 then
        begin
          if (atr and H_FLIP)<>0 then
          begin
            if ((J and $80)<>0)  and ((iPP + 7)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 7] := ((c1 shr 6) and 3) + cc;
            if ((J and $40)<>0) and ((iPP + 6)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 6] := ((c2 shr 6) and 3) + cc;
            if ((J and $20)<>0) and ((iPP + 5)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 5] := ((c1 shr 4) and 3) + cc;
            if ((J and $10)<>0) and ((iPP + 4)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 4] := ((c2 shr 4) and 3) + cc;
            if ((J and $8)<>0) and ((iPP + 3)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 3] := ((c1 shr 2) and 3) + cc;
            if ((J and $4)<>0) and ((iPP + 2)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 2] := ((c2 shr 2) and 3) + cc;
            if ((J and $2)<>0) and ((iPP + 1)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 1] := (c1 and 3) + cc;
            if ((J and $1)<>0) and ((iPP + 0)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 0] := (c2 and 3) + cc;
          end
          else
          begin
            if ((J and $80)<>0) and ((iPP + 0)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 0] := ((c1 shr 6) and 3) + cc;
            if ((J and $40)<>0) and ((iPP + 1)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 1] := ((c2 shr 6) and 3) + cc;
            if ((J and $20)<>0) and ((iPP + 2)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 2] := ((c1 shr 4) and 3) + cc;
            if ((J and $10)<>0) and ((iPP + 3)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 3] := ((c2 shr 4) and 3) + cc;
            if ((J and $8)<>0) and ((iPP + 4)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 4] := ((c1 shr 2) and 3) + cc;
            if ((J and $4)<>0) and ((iPP + 5)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 5] := ((c2 shr 2) and 3) + cc;
            if ((J and $2)<>0) and ((iPP + 6)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 6] := (c1 and 3) + cc;
            if ((J and $1)<>0) and ((iPP + 7)<DXDIB.DIB.WidthBytes) then
              PP[iPP + 7] := (c2 and 3) + cc;
          end;
        end;
        inc(t);
        inc(C, incr);
      until t>=h;
    end;
    dec(n, $04);
  until n<$0000;
end; { RefreshSprite }

procedure TfrmNES.CheckSpriteHit;
var
  vofs : integer;
  HH : byte;
  sprite : ^TSPR;
  y, no, J : byte;
  t, C, vadr : integer;
begin
  vofs := $0;
  sprite := @SRAM[0];
  y := sprite.y;
  no := sprite.no;
  if (PPU[0] and $20)<>0 then { Sprite 16 }
  begin
    vadr := ((no and $1) shl 12) + (no shr 1) * 32;
    HH := 16;
  end
  else
  begin
    if (PPU[0] and $08)<>0 then { Sprite pattern address }
      vofs := $1000;
    vadr := no * 16 + vofs;
    HH := 8;
  end;
  C := vadr;
  HitY := y;
  for t := 0 to (HH - 1) do
  begin
    if t=8 then
      inc(C, 8);
    J := Ord(VPage[C shr 10][C]) or Ord(VPage[(C + 8) shr 10][C + 8]);
    if J<>0 then
      HitY := y + t + 1;
    inc(C);
  end;
end; { CheckSpriteHit }

procedure TfrmNES.FormCreate(Sender: TObject);
begin
  SoundOn := false;
  CPUPeriod := 122;
  RomFilename := '';
  RomPath     := '';
  MRUList := TStringList.Create;
  { inizializza cpu }
  with cpu do
  begin
    A := 0;
    P := 0;
    X := 0;
    Y := 0;
    S := 0;
    PC.W  := 0;
    iPeriod  := CPUPeriod;
    ICount   := 0;
    IRequest := 0;
    AfterCLI := 0;
    IBackup  := 0;
    //void *User;
    TrapBadOps := 0;
    Trap  := 0;
    Trace := 0;
  end;
  { legge le risorse }
  VersionResource.FileName := Application.EXEname;
  Version := VersionResource.VersionInfo.ProductVersion;
  { Read configuration }
  ReadConfig;
  with DXInput1.Keyboard do
  begin
    FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
    AssignKey(KeyAssigns, isUp,      [BUTTON1_UP]);
    AssignKey(KeyAssigns, isDown,    [BUTTON1_DOWN]);
    AssignKey(KeyAssigns, isLeft,    [BUTTON1_LEFT]);
    AssignKey(KeyAssigns, isRight,   [BUTTON1_RIGHT]);
    AssignKey(KeyAssigns, isButton1, [BUTTON1_A]);
    AssignKey(KeyAssigns, isButton2, [BUTTON1_B]);
    AssignKey(KeyAssigns, isButton3, [BUTTON1_START]);
    AssignKey(KeyAssigns, isButton4, [BUTTON1_SELECT]);
  end;
  DXInput1.Joystick.Enabled := BUTTON1_JOY_ENABLED;
  with DXInput2.Keyboard do
  begin
    FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
    AssignKey(KeyAssigns, isUp,      [BUTTON2_UP]);
    AssignKey(KeyAssigns, isDown,    [BUTTON2_DOWN]);
    AssignKey(KeyAssigns, isLeft,    [BUTTON2_LEFT]);
    AssignKey(KeyAssigns, isRight,   [BUTTON2_RIGHT]);
    AssignKey(KeyAssigns, isButton1, [BUTTON2_A]);
    AssignKey(KeyAssigns, isButton2, [BUTTON2_B]);
    AssignKey(KeyAssigns, isButton3, [BUTTON2_START]);
    AssignKey(KeyAssigns, isButton4, [BUTTON2_SELECT]);
  end;
  DXInput2.Joystick.Enabled := BUTTON2_JOY_ENABLED;
  { initialize }
  Randomize;
  Terminal.Lines.Add(VersionResource.VersionInfo.ProductName+ ' ' + Version);
  Terminal.Lines.Add(VersionResource.VersionInfo.FileDescription); //'The NES Emulator');
  Terminal.Lines.Add(VersionResource.VersionInfo.LegalCopyright);  //'(C) 1998 by Dott. Cosimo De Michele');
  Terminal.Lines.Add(VersionResource.VersionInfo.LegalTrademarks);
  Terminal.Lines.Add('');
  { inizializzo variabili }
  CurrentFrame := 1;
  ROM := nil;
  VROM := nil;
  { Creo la palette }
  DXDIB.DIB.BitCount := 8;
  DXDIB.DIB.Width    := NES_SCREEN_WIDTH;
  DXDIB.DIB.Height   := NES_SCREEN_HEIGHT;
end;

procedure TfrmNES.FormDestroy(Sender: TObject);
begin
  WriteConfig;
  MRUList.free;
end;

procedure TfrmNES.mnuExitClick(Sender: TObject);
begin
  Close;
end; { mnuExitClick }

function TfrmNES.MemoryReadByte(Addr: Word): Byte;
begin
  case (Addr shr 13) of
    0 : { $0000 -- RAM }
        Result := RAM[Addr and $7FF];
    1,  { $2000 -- PPU }
    2 : { $4000 -- PSG }
        Result := IORead(Addr);
    3 : { $6000 -- WRAM }
        Result := WRAM[Addr and $1FFF];
  else
    { $8000 -- ROM }
    Result := Ord(Page[Addr shr 13][Addr]);
  end;
end; { MemoryReadByte }

procedure TfrmNES.MemoryWrite(Addr: Word; value: Byte);
begin
  case (Addr shr 13) of
    0 : { $0000 -- RAM }
        RAM[Addr and $7FF] := value;
    1,  { $2000 -- PPU }
    2 : { $4000 -- PSG }
        IOWrite(Addr, value);
    3 : { $6000 -- WRAM }
        WRAM[Addr and $1FFF] := value;
  else
    { $8000 -- ROM }
    MMCWrite(Addr, value);
  end;
end; { MemoryWrite }

function TfrmNES.MemoryReadOpcode(Addr: Word): Byte;
begin
  Result := Ord(Page[Addr shr 13][Addr]);
end; { MemoryReadOpcode }

function  TfrmNES.IORead(A : word): byte;
begin
  case A of
    { PPU }
    $2002 : begin { PPU Status }
              Result := PPU[2];
              PPU[2] := PPU[2] and not(PPU_InVBlank or PPU_SpHit);
            end;
    $2007 : begin
              if PPUADDR_FirstRead then
              begin
                PPUADDR_FirstRead := false;
                Result := NO_DATA;
                Exit;
              end;
              VAddr := VAddr and $3FFF;
              if (VAddr<$2000) then
                Result := Ord(VPage[VAddr shr 10][VAddr])
              else
                Result := VRAM[VAddr];
              if (PPU[0] and $04)<>0 then { INC32 }
                inc(VAddr, 32)
              else
                inc(VAddr);
            end;
    { PGS }
    $4000..$4013, $4015 :
            begin
              Result := PSG[A - $4000];
            end;
    { Joypad }
    $4016 : begin
              Result := ((Joypad1 shr Joypad1Read) and $01);
              Joypad1Read := (Joypad1Read + 1) and 23;
            end;
     $4017 : begin
              Result := (Joypad2 shr Joypad2Read) and $01;
              Joypad2Read := (Joypad2Read + 1) and 23;
            end;
    else
    begin
      Terminal.Lines.Add(format('Ignore read at address $%4x', [A]));
      Result := NO_DATA;
    end;
  end;
end; { IORead }

procedure TfrmNES.IOWrite(A : word; V : byte);
var
  len : integer;
  freq,
  ch : WORD;
begin
  case A of
    $2000 : PPU[0] := V;
    $2001 : PPU[1] := V;
    $2002 : PPU[2] := V;
    $2003 : PPU[3] := V;
    $2004 : begin
              SRAM[PPU[3]] := V;
              inc(PPU[3]);
            end;
    $2005 : begin
              if not(PPUSCROLL_Toggle) then
                PPUSCROLL_X2 := V
              else if V<=239 then
                PPUSCROLL_Y2 := V;
              PPUSCROLL_Toggle := not(PPUSCROLL_Toggle);
              PPUSCROLL_Change := true
            end;
    $2006 : begin
              VAddr := (VAddr shl 8) or V;
              PPUADDR_FirstRead := true;
            end;
    $2007 : begin
              if (PPU[2] and PPU_WriteIgnore)=0 then
              begin
                VAddr := VAddr and $3FFF;
                VRAM[VAddr] := V;
                if ((VAddr and $3000)=$2000) then { Mirroring }
                  VRAM[(VAddr xor MirrorXor) ] := V
                else if (VAddr>=$3F00) and (VAddr<=$3F1F) then
                begin
                   if (VAddr=$3F10) then
                   begin
                     VRAM[$3F00] := V;
                     VRAM[$3F04] := V;
                     VRAM[$3F08] := V;
                     VRAM[$3F0C] := V;
                     VRAM[$3F10] := V;
                     VRAM[$3F14] := V;
                     VRAM[$3F18] := V;
                     VRAM[$3F1C] := V;
                   end;
                   PPU_PalChange := true
                end;
                if (PPU[0] and $04)<>0 then { INC32 }
                  inc(VAddr, 32)
                else
                  inc(VAddr);
              end;
              PPUADDR_FirstRead := false;
            end;
    { PSG }
    $4000,
    $4004,
    $4008,
    $400C : begin
              PSG[A - $4000] := V;
            end;
    $4001 : begin
              PSG[A - $4000] := V;
              SndFreq[0] := (V and $80)<>0 ;
            end;
    $4005 : begin
              PSG[A - $4000] := V;
              SndFreq[1] := (V and $80)<>0;
            end;
    $4009 : begin
              PSG[A - $4000] := V;
              SndFreq[2] := (V and $80)<>0;
            end;
    $400D : begin
              PSG[A - $4000] := V;
              SndFreq[3] := (V and $80)<>0;
            end;
    $4002,
    $4003 : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[0] := ((PSG[$03] and 7) shl 8) or PSG[$02];
              freq := SndFreqVal[0];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$03] shr 3);
              SetFreq(0, freq);
              if (SndON[0]) then
                SetVolume(0, SndVolume)
              else
                SetVolume(0, 0);
              SndLen[0] := (len + 1) * 8;
            end;
    $4006,
    $4007 : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[1] := ((PSG[$07] and 7) shl 8) or PSG[$06];
              freq := SndFreqVal[1];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$07] shr 3);
              SetFreq(1, freq);
              if (SndON[1]) then
                SetVolume(1, SndVolume)
              else
                SetVolume(1, 0);
              SndLen[1] := (len + 1) * 8;
            end;
    $400A,
    $400B : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[2] := ((PSG[$0B] and 7) shl 8) or PSG[$0A];
              freq := SndFreqVal[2];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$0B] shr 3);
              SetFreq(2, freq);
              if (SndON[2]) then
                SetVolume(2, SndVolume)
              else
                SetVolume(2, 0);
              SndLen[2] := len + 2;
            end;
    $400E,
    $400F : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[3] := ((PSG[$0F] and 7) shl 8) or PSG[$0E];
              freq := SndFreqVal[3];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$0F] shr 3);
              SetFreq(3, freq);
              if (len<>0) or (SndON[3]) then
                SetVolume(3, SndVolume)
              else
                SetVolume(3, 0);
              SndLen[3] := len;
            end;
    { PCM }
    $4010,
    $4011,
    $4012,
    $4013 : begin
              PSG[A - $4000] := V;
            end;
    { DMA }
    $4014 : begin
              if V<$80 then
                CopyMemory(@SRAM, @RAM[(V and 7) shl 8], 256)
              else
                CopyMemory(@SRAM, pMemory(Page[V shr 5][V shl 8]), 256)
            end;
    $4015 : begin
              for ch:=0 to 3 do
                if ((V shr ch) and 1)<>0 then
                begin { Enable sound }
                end
                else
                begin { disable sound }
                  SndLen[ch] := 0;
                  SetVolume(ch, 0);
                end;
              PSG[$15] := V;
            end;
    { Joypad }
    $4016 : begin
              if ((PSG[$16] and 1) and not(V and 1))<>0 then
              begin
                Joypad1Read := 0;
                Joypad2Read := 0;
              end;
              PSG[$16] := V;
            end;
    $4017 : begin
              PSG[$17] := V;
            end;
    else
    begin
      Terminal.Lines.Add(format('Ignore write at address $%4x', [A]));
    end;
  end;
end; { IOWrite }

function TfrmNES.cpuLoop : byte;
var
  ch,
  freq : integer;
  freqS : byte;
begin
  { gestione suono }
  dec(SndCount);
  if SndCount<=0 then
  begin
    SndCount := SndPeriod;
    for ch := 0 to 3 do
    begin
      if (SndLen[ch]<>0) and (SndOn[ch]) then
      begin
        if SndFreq[ch] then
        begin
          freqS := (((PSG[ch * 4 + 1] shr 4) and $07)) * ((PSG[ch + 4 + 1] and $07));
          if (PSG[ch * 4 + 1] and $08)<>0 then
            dec(SndFreqVal[ch], freqS)
          else
            inc(SndFreqVal[ch], freqS);
          freq := PSG_BASE div (SndFreqVal[ch] + 1);
          SetFreq(ch, freq and $03FF);
        end;
        dec(SndLen[ch]);
        if SndLen[ch]=0 then
        begin
          SetVolume(ch, 0);
          PSG[$15] := PSG[$15] and not(1 shl ch);
        end;
      end;
    end;
  end;
  { gestione scanline }
  ScanLine := (ScanLine + 1) mod SCANLINEPERFRAME;
  Result := INT_NONE;
  //if (Uperiod=0) and (ScanLine>=0) and (ScanLine<NES_SCREEN_HEIGHT) then
  if (Uperiod=0) and (ScanLine>=MINLINE) and (ScanLine<=MAXLINE) then
    RefreshLine(ScanLine);
  if (ScanLine = 0) then
  begin
    PPU[2] := PPU[2] and not(PPU_InVBlank);
    if Uperiod=0 then
      DrawScreen;
  end;
  if (ScanLine = MAXLINE) then
  begin
    inc(UPeriod);
    UPeriod := UPeriod mod FrameRate;
    if ((PPU[1] and $10)<>0) and (Uperiod = 0) then { SpriteON }
      RefreshSprite(MINLINE);
    PPU[2] := PPU[2] or PPU_InVBlank;
    PPU[2] := PPU[2] and not(PPU_SpHit);
    PSG[$16] := PSG[$16] and $FD;
    PPUSCROLL_Toggle := false;
    if (PPU[0] and $80)<>0 then { VBlank ON }
    begin
      Result := INT_NMI;
      exit;
    end;
  end
  else if (ScanLine = MINLINE) then
  begin
    { Check the Keyboard }
    CheckKb;
    { Check sprites Hit }
    CheckSpriteHit;
    Application.ProcessMessages;
    if not NESisRunning then
    begin
      for ch:= 0 to 3 do
        SetVolume(ch, 0);
      Result := INT_QUIT;
      exit;
    end;
  end;
  if (ScanLine>=MINLINE) and (ScanLine<=MAXLINE) then
  begin
    IRQz(Result);
    if Result<>INT_NONE then
      exit;
    if PPUSCROLL_Change then
    begin
      PPUSCROLL_X1 := PPUSCROLL_X2;
      PPUSCROLL_Y1 := PPUSCROLL_Y2;
      PPUSCROLL_Change := false;
    end;
    if ((PPU[1] and $10)<>0) and (ScanLine=HitY) then { SpriteON }
    begin
      PPU[2] := PPU[2] or PPU_SpHit;
      if (PPU[0] and $40)<>0 then { Sprite Hit ON }
        Result := INT_NMI;
    end;
  end;
end; { cpuLoop }

procedure TfrmNES.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE : if doFullScreen in DXDraw.Options then
                  GoFullScreen
                else
                  Close;
    VK_RETURN : if (ssAlt in Shift) and (UseDirectX) then
                  GoFullScreen;
    VK_F2     : mnuLoadROMClick(mnuLoadROM);
    VK_F3     : mnuFreeROMClick(mnuFreeROM);
  end;
end;

procedure TfrmNES.mnuPauseClick(Sender: TObject);
begin
  if NESisRunning then
    NESisRunning := false
  else
  begin
    NESisRunning := true;
    Run6502(cpu);
  end;
end;

procedure TfrmNES.mnuResetClick(Sender: TObject);
begin
  NESisRunning := false;
  ResetNes;
  NESisRunning := true;
  Run6502(cpu);
end;

procedure TfrmNES.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  mnuFreeROM.Click;
end;

procedure TfrmNES.mnuRomClick(Sender: TObject);
var
  frmROMProperty: TfrmROMProperty;
begin
 frmROMProperty := TfrmROMProperty.Create(Self);
 with frmROMProperty do
 begin
   edFilename.Text := UpperCase(RomFilename);
   edHeader.Text   := RomHeader.RomType;
   edProgPages.Text:= IntToStr(Ord(RomHeader.PRGROM));
   lblROM.Caption  := format('[%d Kb]', [Ord(RomHeader.PRGROM) * 16]);
   edCharPages.Text:= IntToStr(Ord(RomHeader.CHRRAM));
   lblVROM.Caption := format('[%d Kb]', [Ord(RomHeader.CHRRAM) * 8]);
   edMapping.Text  := '#' + IntToStr(MMCType and 15) + ', ' + Mappers;
   cbVertMirror.Checked := (Mirroring = 0);
   cbBattery.Checked    := (ROMType and ROM_WRAM)<>0;
   cbTrainer.Checked    := (ROMType and ROM_TRAINER)<>0;
   cbFourScreen.Checked := (ROMType and ROM_FOUR)<>0;
   edExpansion.Text := ROMHeader.ROMctrl2 + RomHeader.Expansion;
   ShowModal;
 end;
end;

procedure TfrmNES.mnuAboutClick(Sender: TObject);
var
  frmAbout: TfrmAbout;
begin
  frmAbout := TfrmAbout.Create(Application);
  frmAbout.ShowModal;
  frmAbout.free;
end;

procedure TfrmNES.mnuFreeROMClick(Sender: TObject);
begin
  mnuRom.enabled := false;
  mnuFreeROM.enabled := false;
  mnuDebug.enabled := false;
  mnuSaveSnapShot.enabled := false;
  TrashMachine;
end;

procedure TfrmNES.mnuSettingClick(Sender: TObject);
var
  frmSetting: TfrmSetting;
begin
  frmSetting := TfrmSetting.Create(Self);
  frmSetting.ShowModal;
  if frmSetting.ModalResult = mrOK then
  begin
    { Joypad 1 }
    with DXInput1.Keyboard do
    begin
      FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
      AssignKey(KeyAssigns, isUp,      [BUTTON1_UP]);
      AssignKey(KeyAssigns, isDown,    [BUTTON1_DOWN]);
      AssignKey(KeyAssigns, isLeft,    [BUTTON1_LEFT]);
      AssignKey(KeyAssigns, isRight,   [BUTTON1_RIGHT]);
      AssignKey(KeyAssigns, isButton1, [BUTTON1_A]);
      AssignKey(KeyAssigns, isButton2, [BUTTON1_B]);
      AssignKey(KeyAssigns, isButton3, [BUTTON1_START]);
      AssignKey(KeyAssigns, isButton4, [BUTTON1_SELECT]);
    end;
    DXInput1.Joystick.Enabled := BUTTON1_JOY_ENABLED;
    { Joypad 2 }
    with DXInput2.Keyboard do
    begin
      FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
      AssignKey(KeyAssigns, isUp,      [BUTTON2_UP]);
      AssignKey(KeyAssigns, isDown,    [BUTTON2_DOWN]);
      AssignKey(KeyAssigns, isLeft,    [BUTTON2_LEFT]);
      AssignKey(KeyAssigns, isRight,   [BUTTON2_RIGHT]);
      AssignKey(KeyAssigns, isButton1, [BUTTON2_A]);
      AssignKey(KeyAssigns, isButton2, [BUTTON2_B]);
      AssignKey(KeyAssigns, isButton3, [BUTTON2_START]);
      AssignKey(KeyAssigns, isButton4, [BUTTON2_SELECT]);
    end;
    DXInput2.Joystick.Enabled := BUTTON2_JOY_ENABLED;
  end;
  frmSetting.free;
end;

procedure TfrmNES.ReadConfig;
var
  Reg : TRegistry;
begin
  Reg := TRegistry.Create;
  if Reg.OpenKey(RegKey, true) then
  begin
    { NES }
    if Reg.ValueExists('FrameRate') then
      FrameRate := Reg.ReadInteger('FrameRate')
    else
      FrameRate := 3;
    if Reg.ValueExists('ViewFrequency') then
      ViewFreq := Reg.ReadBool('ViewFrequency')
    else
      ViewFreq := false;
    { Sound }
    if Reg.ValueExists('SoundVolume') then
      SndVolume := Reg.ReadInteger('SoundVolume')
    else
      SndVolume := 10;
    if Reg.ValueExists('SoundChannel1') then
      SndON[0] := Reg.ReadBool('SoundChannel1')
    else
      SndON[0] := true;
    if Reg.ValueExists('SoundChannel2') then
      SndON[1] := Reg.ReadBool('SoundChannel2')
    else
      SndON[1] := true;
    if Reg.ValueExists('SoundChannel3') then
      SndON[2] := Reg.ReadBool('SoundChannel3')
    else
      SndON[2] := true;
    if Reg.ValueExists('SoundChannel4') then
      SndON[3] := Reg.ReadBool('SoundChannel4')
    else
      SndON[3] := true;
    { Screen }
    if Reg.ValueExists('WIN') then
      WIN_MODE := Reg.ReadInteger('WIN')
    else
      WIN_MODE := 0; // Normal Screen
    if Reg.ValueExists('SCREEN') then
      SCREEN_MODE := Reg.ReadString('SCREEN')
    else
    begin
      SCREEN_MODE := ''; // Default Screen
      WIN_MODE := 0; // Normal Screen
    end;
    { Most Recent Used }
    MRUList.Add(Reg.ReadString('MRU1'));
    MRUList.Add(Reg.ReadString('MRU2'));
    MRUList.Add(Reg.ReadString('MRU3'));
    MRUList.Add(Reg.ReadString('MRU4'));
    MRUList.Add(Reg.ReadString('MRU5'));
  end;
  Reg.CloseKey;
  if Reg.OpenKey(RegKey + '\JOY1', true) then
  begin
    { Joypad 1 }
    if Reg.ValueExists('ENABLED') then
      BUTTON1_JOY_ENABLED := Reg.ReadBool('ENABLED')
    else
      BUTTON1_JOY_ENABLED := false;
    if Reg.ValueExists('UP') then
      BUTTON1_UP := Reg.ReadInteger('UP')
    else
      BUTTON1_UP := VK_UP;
    if Reg.ValueExists('DOWN') then
      BUTTON1_DOWN := Reg.ReadInteger('DOWN')
    else
      BUTTON1_DOWN := VK_DOWN;
    if Reg.ValueExists('LEFT') then
      BUTTON1_LEFT := Reg.ReadInteger('LEFT')
    else
      BUTTON1_LEFT := VK_LEFT;
    if Reg.ValueExists('RIGHT') then
      BUTTON1_RIGHT := Reg.ReadInteger('RIGHT')
    else
      BUTTON1_RIGHT := VK_RIGHT;
    if Reg.ValueExists('A') then
      BUTTON1_A := Reg.ReadInteger('A')
    else
      BUTTON1_A := VK_SPACE;
    if Reg.ValueExists('B') then
      BUTTON1_B := Reg.ReadInteger('B')
    else
      BUTTON1_B := VK_CONTROL;
    if Reg.ValueExists('START') then
      BUTTON1_START := Reg.ReadInteger('START')
    else
      BUTTON1_START := VK_RETURN;
    if Reg.ValueExists('SELECT') then
      BUTTON1_SELECT := Reg.ReadInteger('SELECT')
    else
      BUTTON1_SELECT := VK_TAB;
  end;
  Reg.CloseKey;
  if Reg.OpenKey(RegKey + '\JOY2', true) then
  begin
    { Joypad 2 }
    if Reg.ValueExists('ENABLED') then
      BUTTON2_JOY_ENABLED := Reg.ReadBool('ENABLED')
    else
      BUTTON2_JOY_ENABLED := false;
    if Reg.ValueExists('UP') then
      BUTTON2_UP := Reg.ReadInteger('UP')
    else
      BUTTON2_UP := Ord('E');
    if Reg.ValueExists('DOWN') then
      BUTTON2_DOWN := Reg.ReadInteger('DOWN')
    else
      BUTTON2_DOWN := Ord('X');
    if Reg.ValueExists('LEFT') then
      BUTTON2_LEFT := Reg.ReadInteger('LEFT')
    else
      BUTTON2_LEFT := Ord('S');
    if Reg.ValueExists('RIGHT') then
      BUTTON2_RIGHT := Reg.ReadInteger('RIGHT')
    else
      BUTTON2_RIGHT := Ord('D');
    if Reg.ValueExists('A') then
      BUTTON2_A := Reg.ReadInteger('A')
    else
      BUTTON2_A := VK_SHIFT;
    if Reg.ValueExists('B') then
      BUTTON2_B := Reg.ReadInteger('B')
    else
      BUTTON2_B := Ord('Z');
    if Reg.ValueExists('START') then
      BUTTON2_START := Reg.ReadInteger('START')
    else
      BUTTON2_START := Ord('Q');
    if Reg.ValueExists('SELECT') then
      BUTTON2_SELECT := Reg.ReadInteger('SELECT')
    else
      BUTTON2_SELECT := Ord('W');
  end;
  Reg.CloseKey;
  Reg.free;
end; { ReadConfig }

procedure TfrmNES.WriteConfig;
var
  Reg : TRegistry;
  i : integer;
begin
  Reg := TRegistry.Create;
  if Reg.OpenKey(RegKey, true) then
  begin
    { NES }
    Reg.WriteInteger('FrameRate', FrameRate);
    Reg.WriteBool('ViewFrequency', ViewFreq);
    { Sound }
    Reg.WriteInteger('SoundVolume', SndVolume);
    for i:=0 to 3 do
      Reg.WriteBool('SoundChannel' + IntToStr(i + 1), SndON[i]);
    { Screen }
    Reg.WriteString('SCREEN', SCREEN_MODE);
    Reg.WriteInteger('WIN', WIN_MODE);
    { Most recently used }
    for i:=0 to (MRUCount-1) do
      Reg.WriteString(format('MRU%d', [i + 1]), MRUList.Strings[i]);
  end;
  Reg.CloseKey;
  if Reg.OpenKey(RegKey + '\JOY1', true) then
  begin
    { Joypad 1 }
    Reg.WriteInteger('UP', BUTTON1_UP);
    Reg.WriteInteger('DOWN', BUTTON1_DOWN);
    Reg.WriteInteger('LEFT', BUTTON1_LEFT);
    Reg.WriteInteger('RIGHT', BUTTON1_RIGHT);
    Reg.WriteInteger('A', BUTTON1_A);
    Reg.WriteInteger('B', BUTTON1_B);
    Reg.WriteInteger('START', BUTTON1_START);
    Reg.WriteInteger('SELECT', BUTTON1_SELECT);
    Reg.WriteBool('ENABLED', BUTTON1_JOY_ENABLED);
  end;
  Reg.CloseKey;
  if Reg.OpenKey(RegKey + '\JOY2', true) then
  begin
    { Joypad 2 }
    Reg.WriteInteger('UP', BUTTON2_UP);
    Reg.WriteInteger('DOWN', BUTTON2_DOWN);
    Reg.WriteInteger('LEFT', BUTTON2_LEFT);
    Reg.WriteInteger('RIGHT', BUTTON2_RIGHT);
    Reg.WriteInteger('A', BUTTON2_A);
    Reg.WriteInteger('B', BUTTON2_B);
    Reg.WriteInteger('START', BUTTON2_START);
    Reg.WriteInteger('SELECT', BUTTON2_SELECT);
    Reg.WriteBool('ENABLED', BUTTON2_JOY_ENABLED);
  end;
  Reg.CloseKey;
  Reg.free;
end; { WriteConfig }

procedure TfrmNES.InitScreen;
begin
  ScreenWidth  := NES_SCREEN_WIDTH;
  ScreenHeight := NES_SCREEN_HEIGHT;
 try
    DXDraw.Initialize;
    UseDirectX := true;
  except
    UseDirectX := false;
  end;
  //UseDirectX := false;
  { inizializzo video }
  DXDraw.SurfaceWidth  := ScreenWidth;
  DXDraw.SurfaceHeight := ScreenHeight;
  if (WIN_MODE=1) and (UseDirectX) then
    GoFullScreen;
  Terminal.visible := false;
  DXDraw.Visible := UseDirectX;
  { Play Logo }
  PlayLogo;
end; { InitScreen }

procedure TfrmNES.TrashScreen;
begin
  DXDraw.Finalize;
  Terminal.visible := true;
  DXDraw.visible := false;
  Menu := MainMenu;
  StatusBar.visible := true;
  BorderStyle := bsSingle;
  DXDraw.Cursor := crDefault;
  DXDraw.Options := DXDraw.Options - [doFullScreen];
  if WIN_MODE=1 then
    RestoreWindow;
end; { TrashScreen }

procedure TfrmNES.GoFullScreen;
var
  s : string;
  i, AWidth, AHeight, ABitCount: Integer;
begin
  DXDraw.Finalize;
  if doFullScreen in DXDraw.Options then
  begin { from FullScreen to normal screen }
    RestoreWindow;
    Menu := MainMenu;
    StatusBar.visible := true;
    BorderStyle := bsSingle;
    DXDraw.Cursor := crDefault;
    DXDraw.Options := DXDraw.Options - [doFullScreen];
  end
  else
  begin { from normal screen to FullScreen }
    StoreWindow;
    Menu := nil;
    StatusBar.visible := false;
    BorderStyle := bsNone;
    DXDraw.Cursor := crNone;
    if SCREEN_MODE<>'' then
    begin
      s := SCREEN_MODE;
      i := Pos('x', s);
      AWidth := StrToInt(Copy(s, 1, i-1));
      s := Copy(s, i+1, Length(s));
      i := Pos('x', s);
      AHeight := StrToInt(Copy(s, 1, i-1));
      s := Copy(s, i+1, Length(s));
      ABitCount := StrToInt(s);
      DXDraw.Display.Width := AWidth;
      DXDraw.Display.Height := AHeight;
      DXDraw.Display.BitCount := ABitCount;
    end;
    DXDraw.Options := DXDraw.Options + [doFullScreen];
  end;
  DXDraw.Initialize;
end;

procedure TfrmNES.mnuSavesnapshotClick(Sender: TObject);
begin
  if SaveDialog.Execute then
  begin
    RefreshSprite(MINLINE);
    DXDIB.DIB.SaveToFile(SaveDialog.Filename);
    Terminal.Lines.Add('Snapshot saved as "' + ExtractFileName(SaveDialog.Filename) + '".');
  end;
end;

procedure TfrmNES.MRUClick(Sender: TObject);
var
  index : integer;
  ItemClicked : TMenuItem;
begin
  ItemClicked := TMenuItem(Sender);
  index := ItemClicked.MenuIndex - MRU1.MenuIndex;
  RunMachine(MRUList.Strings[index]);
end;

procedure TfrmNES.mnuFileClick(Sender: TObject);
var
  i, index : integer;
begin
  index := MRU1.MenuIndex;
  for i:=0 to (MRUCount - 1) do
  begin
    mnuFile.Items[index + i].visible := (MRUList.Strings[i]<>'');
    mnuFile.Items[index + i].Caption := MRUList.Strings[i];
  end;
  MRUSeparator.visible := mnuFile.Items[index].visible;
end;

procedure TfrmNES.mnuCPUClick(Sender: TObject);
begin
  mnuReset.Enabled := RomFilename<>'';
  mnuPause.Enabled := RomFilename<>'';
  if NESisRunning then
    mnuPause.Caption := '&Pause'
  else
    mnuPause.Caption := '&Run';
end; { mnuCPUClick }

procedure TfrmNES.StatusBarDrawPanel(StatusBar: TStatusBar;
  Panel: TStatusPanel; const Rect: TRect);
var
  bmp: TBitmap;
  g: TGraphic;
begin
  if NesisRunning then
  begin
    bmp := TBitmap.Create;
    bmp.Width := 16;
    bmp.Height := 16;
    if UseDirectX then
      g := DXImageList.Items[2].Picture.Graphic
    else
      g := DXImageList.Items[3].Picture.Graphic;
    bmp.Canvas.Draw(0, 0, g);
    bmp.Transparent := true;
    StatusBar.Canvas.Draw(Rect.Left, Rect.Top, bmp);
    if SoundOn then
      g := DXImageList.Items[4].Picture.Graphic
    else
      g := DXImageList.Items[5].Picture.Graphic;
    bmp.Canvas.Draw(0, 0, g);
    bmp.Transparent := true;
    StatusBar.Canvas.Draw(Rect.Left + 18, Rect.Top, bmp);
    bmp.free;
  end;
end;

procedure TfrmNES.mnuDebugClick(Sender: TObject);
var
  frmDebug: TfrmDebug;
  NESwasRunning: boolean;
begin
  NESwasRunning := NESisRunning;
  if NESisRunning then
    NESisRunning := false;
  frmDebug := TfrmDebug.Create(Self);
  try
    frmDebug.ShowModal;
  finally
    frmDebug.free;
  end;
  if NESwasRunning then
  begin
    NESisRunning := true;
    Run6502(cpu);
  end;
end;

end.
